/*
    Clickable.js
    Copyright (c) 2009-2013 musictheory.net, LLC.  All rights reserved.
*/

var clickable = (function() { "use strict";

var _idToStorageMap   = { };
var _idToCallbacksMap = { };
var _nextID           = 0;
var _activeID         = 0;
var _didInitialize    = false;
var _prefersTouchEvents = undefined;
var _isScrolling = false;

function prefersTouchEvents()
{
    if (_prefersTouchEvents === undefined) {
        var ua = navigator.userAgent;

        // Check for iPhone OS devices, MobileSafari
        var yn = /(iPad|iPhone|iPod).*Apple.*Safari/.test(ua);

        // Check for iPhone OS, other WebKit view
        if (!yn) {
            yn = /(iPad|iPhone|iPod).*AppleWebKit/.test(ua);
        }

        // Now check for Android 1.5 or later
        if (!yn) {
            var matches = ua.match(/Android\s+([0-9.]+)/);
            yn = !!matches && (parseFloat(matches[1]) >= 1.5);
        }

        if (yn) {
            yn = ("ontouchstart" in document.documentElement);
        }

        _prefersTouchEvents = yn;
    }
    return _prefersTouchEvents;
}

function _initialize()
{
    if (prefersTouchEvents()) {
        $(document.body).on({
            "touchstart touchmove": _handleTouchOn,
            "touchcancel touchend": _handleTouchOff
        });

    } else {
        $(window).on("unload mouseup mousedown", _clearContext);
    }

    _didInitialize = true;
}


function _clearContext(e)
{
    if (_activeID) {
        _update(_activeID, null, function(storage) {
            storage.raw.down = false;
            storage.raw.touchCanceled = false;
        });

        _activeID = 0;
    }
}


function _getStorage(id)
{
    return _idToStorageMap[id];
}


function _getCallbacks(id)
{
    return _idToCallbacksMap[id];
}


function _stop(e)
{
    e.stopPropagation();
    e.preventDefault();
}


function _updateState(storage)
{
    var raw         = storage.raw;
    var enabled     = raw.enabled;
    var hover       = enabled && (raw.inside || (!!raw.isTouch && raw.down));
    var highlighted = enabled && hover && raw.down && (storage.id == _activeID);

    var change = !storage.state ||
        (storage.state.enabled     != enabled) ||
        (storage.state.highlighted != highlighted) ||
        (storage.state.hover       != hover);

    if (change) {
        var c = storage.classes;
        var $element = $(storage.element);

        if (c.disabled)    $element.toggleClass(c.disabled,   !enabled);
        if (c.highlighted) $element.toggleClass(c.highlighted, highlighted);
        if (c.hover)       $element.toggleClass(c.hover,       hover);
    }

    storage.state.enabled     = enabled;
    storage.state.highlighted = highlighted;
    storage.state.hover       = hover;

    return change;
}


function _update(id, e, callback, touch)
{
    var storage  = _getStorage(id);
    if (!storage) return;

    var callbacks = _getCallbacks(id);

    var oldActive = _activeID;
    var oldDown   = storage.raw.down;
    if (callback) callback(storage);
    var newActive = _activeID;
    var newDown   = storage.raw.down;

    if (storage.state.enabled) {
        if (oldDown && !newDown) {
            if (callbacks.up) {
                callbacks.up(e, storage);
            }

            if (!storage.raw.willActivate && callbacks.cancel) {
                callbacks.cancel(e, storage);
            }

        }

        if (!oldDown && newDown && callbacks.down) {
            callbacks.down(e, storage);
        }
    }    

    if (oldActive != id) {
        _update(oldActive);
    }

    if (_updateState(storage)) {
        if (callbacks.update) callbacks.update(e, storage, storage.state);
    }

    if (newActive != id) {
        _update(newActive);
    }
}


function _setupMouse(id, $element, options, storage, callbacks)
{
    var raw = storage.raw;

    function onMove(e) {
        if (storage.state.enabled) {
            if (callbacks.move) {
                callbacks.move(e, storage);
            }
        }
    }

    $element.on("mousedown", function(e) {
        _update(id, e, function() {
            raw.down = true; 
            _activeID = id;
        });

        _stop(e);

        $element.on("mousemove", onMove);
    });

    $element.on("mouseup", function(e) {
        if (_activeID == id) {
            _stop(e);
        }

        _update(id, e, function() {
            raw.down = false;
            if (_activeID == id) _activeID = 0;
        });

        $element.off("mousemove", onMove);
    });

    $element.on("mouseenter", function(e) {
        _update(id, e, function() { raw.inside = true; });

        if (_activeID == id && callbacks.move) {
            $element.on("mousemove", onMove);
        }
    });

    $element.on("mouseleave", function(e) {
        _update(id, e, function() { raw.inside = false; });
        if (callbacks.move) $element.off("mousemove", onMove);
    });

    $element.on("click", function(e) {
        raw.willActivate = true;

        if (storage.state.enabled) {
            if (callbacks.activate) {
                callbacks.activate(e, storage);
            }

            _stop(e);
        }

        _update(id, e, function() { 
            if (_activeID !== id) _activeID = 0;
        });

        raw.willActivate = false;

        $element.off("mousemove", onMove);
    })
}




function _getNativeEvent(e)
{
    while (e && typeof e.originalEvent !== "undefined") {
        e = e.originalEvent;
    }

    return e;
}


function _handleTouchOn(e)
{
    if (!_activeID) return;

    var storage = _getStorage(_activeID);
    if (!storage) return;

    var raw = storage.raw;
    if (!raw) return;

    var touches = _getNativeEvent(e).touches;

    if (touches.length > 1) {
        _update(_activeID, e, function() {
            raw.touchCanceled = true;
            raw.down = false;
        }); 

    } else if (touches.length == 1) {
        var touch = touches[0];

        if (e.pageX   === undefined) e.pageX   = touch.pageX;
        if (e.pageY   === undefined) e.pageY   = touch.pageY;
        if (e.screenX === undefined) e.screenX = touch.screenX;
        if (e.screenY === undefined) e.screenY = touch.screenY;

        var isMove = e.type == "touchmove";
        var tooFar = isMove && (Math.sqrt(
            Math.pow(touch.screenX - raw.touchStartX, 2) +
            Math.pow(touch.screenY - raw.touchStartY, 2)
        ) > 10);

        if (tooFar || raw.touchCanceled) {
            _update(_activeID, e, function() {
                raw.touchCanceled = true;
                raw.down = false;
            });

        } else {
            _update(_activeID, e, function() {
                if (!isMove) {
                    raw.touchStartX = touch.screenX;
                    raw.touchStartY = touch.screenY;
                }

                raw.down = true;
                raw.touchScreenX  = touch.screenX;
                raw.touchScreenY  = touch.screenY;
                raw.touchPageX    = touch.pageX;
                raw.touchPageY    = touch.pageY;
                raw.touchCanceled = false;
                raw.isTouch       = true;
            });
        }
    }
}


function _handleTouchOff(e)
{
    if (!_activeID) return;

    var storage   = _getStorage(_activeID);
    var callbacks = _getCallbacks(_activeID);
    var raw       = storage && storage.raw;
    var touches   = _getNativeEvent(e).touches;

    if (storage && callbacks && raw) {
        if (e.type == "touchcancel") {
            _update(_activeID, e, function() {
                raw.touchCanceled = true;
                raw.down = false;
            });
        }

        if (storage.state.enabled && !raw.touchCanceled) {
            e.pageX   = raw.touchPageX;
            e.pageY   = raw.touchPageY;
            e.screenX = raw.touchScreenX;
            e.screenY = raw.touchScreenY;

            // Dispatch cancel/up event before activate 
            _update(_activeID, e, function() {
                raw.touchCanceled = true;
                raw.down = false;
            });

            if (callbacks.activate) {
                raw.willActivate = true;
                callbacks.activate(e, storage);
                raw.willActivate = false;
            }
        }
    }

    if (!touches.length) {
        _clearContext(e);
    }

    return false;
}


function _setupTouch(id, $element, options, storage)
{
    $element.on("touchstart", function(e) {
        var touches = _getNativeEvent(e).touches;

        if (!_activeID && touches.length == 1 && !_isScrolling) {
            _activeID = id;
        }
    });
}


function make(element, options)
{
    if (!_didInitialize) {
        _initialize();
    }

    var id = ++_nextID;

    var $element = $(element);

    var storage = {
        id: id,
        element: $element[0],
        raw: {
            enabled: true,
            down: false,
            inside: false,
            willActivate: false
        },
        state: {
            enabled: true,
            highlighted: false,
            hover: false
        }
    };

    if (options.info) storage.info = storage.info;
    if (options.tag)  storage.tag  = options.tag;

    var callbacks = { };
    if (options.activate) callbacks.activate = options.activate;
    if (options.click)    callbacks.activate = options.click;
    if (options.update)   callbacks.update   = options.update;
    if (options.down)     callbacks.down     = options.down;
    if (options.up)       callbacks.up       = options.up;
    if (options.cancel)   callbacks.cancel   = options.cancel;
    if (options.move)     callbacks.move     = options.move;

    storage.classes = options.classes || { };
    if (options.activeClass) storage.classes.highlighted = options.activeClass;

    _idToStorageMap[id]   = storage;
    _idToCallbacksMap[id] = callbacks;

    if (prefersTouchEvents()) {
        _setupTouch(id, $element, options, storage);
    } else {
        _setupMouse(id, $element, options, storage, callbacks);
    }

    return id;
}

    
function setEnabled(id, enabled)
{
    var storage = _getStorage(id);

    if (storage) {
        _update(id, null, function(storage) {
            storage.raw.enabled = enabled;
        });
    }
}

function setIsScrolling(isScrolling)
{
    _isScrolling = isScrolling;
}

// we use the 'clickable' class so that controls unhighlight when the touch moves
function installClickablesForCSSSelector(selector) {
  var ids = [];
  $(selector).each(function() {
    var element = this;
    var f = element.onclick;

    if (f) {
        element.onclick = "";

        ids.push(clickable.install(this, {
            activate: function(e) {
                f.apply(element, [ e ]);
            },
            classes: { highlighted: "highlighted" }
        }));
    }
  });
  return ids;
}

return {
    make:       make,
    install:    make,
    installClickablesForCSSSelector: installClickablesForCSSSelector,
    setEnabled: setEnabled,
    setIsScrolling: setIsScrolling,
}


})();



